package verifier;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.ConcreteCrypto.ElGamalCiphertext;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import prover.ProofOrganizer;
import qilin.primitives.RandomOracle;
import qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.List;


public class Verifier implements Mix2ZeroKnowledgeVerifier {


    /**
     * Return true iff the proof is valid.
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @return
     */

    ECGroup group;
    RandomOracle randomOracle;
    ECElGamalEncryption encryptor;
    ECPoint g,h;

    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.encryptor = encryptor;
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
    }

    public BigInteger hash(Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle input) {
        byte[] arr = input.toByteArray();
        return new BigInteger(this.randomOracle.hash(arr,arr.length));
    }

    private ECPoint convert2ECPoint(ByteString bs){
        return group.decode(bs.toByteArray());
    }


    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof) throws InvalidProtocolBufferException {

        ProofOrganizer organizer = new ProofOrganizer(in1,in2,out1,out2);
        return verifyOrProof(organizer.getE1(ProofOrganizer.OrProofOrder.first),
                organizer.getE2(ProofOrganizer.OrProofOrder.first),
                organizer.getE1New(ProofOrganizer.OrProofOrder.first),
                organizer.getE2New(ProofOrganizer.OrProofOrder.first),
                proof.getFirst(),ProofOrganizer.OrProofOrder.first)&&
                verifyOrProof(organizer.getE1(ProofOrganizer.OrProofOrder.second),
                        organizer.getE2(ProofOrganizer.OrProofOrder.second),
                        organizer.getE1New(ProofOrganizer.OrProofOrder.second),
                        organizer.getE2New(ProofOrganizer.OrProofOrder.second),
                        proof.getSecond(),ProofOrganizer.OrProofOrder.second)&&
                verifyOrProof(organizer.getE1(ProofOrganizer.OrProofOrder.third),
                        organizer.getE2(ProofOrganizer.OrProofOrder.third),
                        organizer.getE1New(ProofOrganizer.OrProofOrder.third),
                        organizer.getE2New(ProofOrganizer.OrProofOrder.third),
                        proof.getThird(),ProofOrganizer.OrProofOrder.third)&&
                verifyOrProof(organizer.getE1(ProofOrganizer.OrProofOrder.fourth),
                        organizer.getE2(ProofOrganizer.OrProofOrder.fourth),
                        organizer.getE1New(ProofOrganizer.OrProofOrder.fourth),
                        organizer.getE2New(ProofOrganizer.OrProofOrder.fourth),
                        proof.getFourth(),ProofOrganizer.OrProofOrder.fourth);
    }



    public boolean verifyOrProof(Crypto.RerandomizableEncryptedMessage e1,
                                 Crypto.RerandomizableEncryptedMessage e2,
                                 Crypto.RerandomizableEncryptedMessage e1Tag,
                                 Crypto.RerandomizableEncryptedMessage e2Tag,
                                 Mixing.ZeroKnowledgeProof.OrProof orProof, ProofOrganizer.OrProofOrder orProofOrder) throws InvalidProtocolBufferException {

        ElGamalCiphertext e1ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1);
        ElGamalCiphertext e2ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2);
        ElGamalCiphertext e1TagElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1Tag);
        ElGamalCiphertext e2TagElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2Tag);
        return verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,orProof,orProofOrder);
    }

    public boolean verifyElGamaOrProof(ElGamalCiphertext e1,
                                 ElGamalCiphertext e2,
                                 ElGamalCiphertext e1New,
                                 ElGamalCiphertext e2New,
                                 Mixing.ZeroKnowledgeProof.OrProof orProof, ProofOrganizer.OrProofOrder orProofOrder)
    {

        Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle forRandomOracle =
                Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle.newBuilder()
                        .setG1(orProof.getG1())
                        .setH1(orProof.getH1())
                        .setG2(orProof.getG2())
                        .setH2(orProof.getH2())
                        .setG1Tag(orProof.getG1Tag())
                        .setH1Tag(orProof.getH1Tag())
                        .setG2Tag(orProof.getG2Tag())
                        .setH2Tag(orProof.getH2Tag())
                        .setU(orProof.getU())
                        .setV(orProof.getV())
                        .setUTag(orProof.getUTag())
                        .setVTag(orProof.getVTag())
                        .build();

        ECPoint g1,g2,h1,h2;
        ECPoint g1Tag,g2Tag,h1Tag,h2Tag;
        g1 = convert2ECPoint(orProof.getG1());
        g2 = convert2ECPoint(orProof.getG2());
        h1 = convert2ECPoint(orProof.getH1());
        h2 = convert2ECPoint(orProof.getH2());
        g1Tag = convert2ECPoint(orProof.getG1Tag());
        g2Tag = convert2ECPoint(orProof.getG2Tag());
        h1Tag = convert2ECPoint(orProof.getH1Tag());
        h2Tag = convert2ECPoint(orProof.getH2Tag());

        ECPoint u,v,uTag,vTag;
        u = convert2ECPoint(orProof.getU());
        v = convert2ECPoint(orProof.getV());
        uTag = convert2ECPoint(orProof.getUTag());
        vTag = convert2ECPoint(orProof.getVTag());

        BigInteger c1,c2,z,zTag;
        c1 = new BigInteger(orProof.getC1().toByteArray());
        c2 = new BigInteger(orProof.getC2().toByteArray());
        z = new BigInteger(orProof.getZ().toByteArray());
        zTag = new BigInteger(orProof.getZTag().toByteArray());

        ECPoint[] given = new ECPoint[12];
        ECPoint[] expected = new ECPoint[12];
        String[] description = new String[12];
        int i = 0;
        given[i++] = g1;
        given[i++] = h1;
        given[i++] = g2;
        given[i++] = h2;
        given[i++] = g1Tag;
        given[i++] = h1Tag;
        given[i++] = g2Tag;
        given[i++] = h2Tag;
        given[i++] = group.multiply(g1, z);
        given[i++] = group.multiply(g2, z);
        given[i++] = group.multiply(g1Tag, zTag);
        given[i] = group.multiply(g2Tag, zTag);
        i = 0;
        expected[i++] = g;
        expected[i++] = group.add(convert2ECPoint(e1New.getC1()), group.negate(convert2ECPoint(e1.getC1())));
        expected[i++] = h;
        expected[i++] = group.add(convert2ECPoint(e1New.getC2()), group.negate(convert2ECPoint(e1.getC2())));
        expected[i++] = g;
        expected[i++] = group.add(convert2ECPoint(e2New.getC1()), group.negate(convert2ECPoint(e2.getC1())));
        expected[i++] = h;
        expected[i++] = group.add(convert2ECPoint(e2New.getC2()), group.negate(convert2ECPoint(e2.getC2())));
        expected[i++] = group.add(u, group.multiply(h1,c1));
        expected[i++] = group.add(v, group.multiply(h2,c1));
        expected[i++] = group.add(uTag, group.multiply(h1Tag,c2));
        expected[i] = group.add(vTag, group.multiply(h2Tag,c2));
        i = 0;
        description[i++] = "g1 != g";
        description[i++] = "h1 != e1New.c1/e1.c1";
        description[i++] = "g2 != h";
        description[i++] = "h2 != e1New.c2/e1.c2";
        description[i++] = "g1Tag != g";
        description[i++] = "h1Tag != e2New.c1/e2.c1";
        description[i++] = "g2Tag != h";
        description[i++] = "h2Tag != e2New.c2/e2.c2";
        description[i++] = "g1 ^ z != u * ( h1 ^ c1 )";
        description[i++] = "g2 ^ z != v * ( h2 ^ c1 )";
        description[i++] = "g1Tag ^ zTag != uTag * ( h1Tag ^ c2 )";
        description[i] = "g2Tag ^ zTag != vTag * ( h2Tag ^ c2 )";
        boolean first = true;
        for (int j = 0; j < given.length ; j++){
            if(!given[j].equals(expected[j])){
                if (first){
                    first = false;
                    System.out.print("\n\n\n");
                    System.out.println(orProofOrder.toString());
                }
                System.out.println(description[j]);
            }
        }

        return   //input
                 g1.equals(g)&&
                 h1.equals(group.add(convert2ECPoint(e1New.getC1())
                         , group.negate(convert2ECPoint(e1.getC1()))))&&
                 g2.equals(h)&&
                 h2.equals(group.add(convert2ECPoint(e1New.getC2())
                         , group.negate(convert2ECPoint(e1.getC2()))))&&
                 // input'
                 g1Tag.equals(g)&&
                 h1Tag.equals(group.add(convert2ECPoint(e2New.getC1())
                         , group.negate(convert2ECPoint(e2.getC1()))))&&
                 g2Tag.equals(h)&&
                 h2Tag.equals(group.add(convert2ECPoint(e2New.getC2())
                         , group.negate(convert2ECPoint(e2.getC2())))) &&
                 // hash
                 // assert (c1 + c2 ) % group size == hash (imput + step1) % group size
                 new BigInteger(orProof.getC1().toByteArray()).add(new BigInteger(orProof.getC2().toByteArray())).mod(group.orderUpperBound())
                 .equals(hash(forRandomOracle).mod(group.orderUpperBound()).mod(group.orderUpperBound()))&&
                 // proof
                 // g1 ^ z == u * ( h1 ^ c1 ) && g2 ^ z == v * ( h2 ^ c1 ) && the same for tag case
                 group.multiply(g1, z).equals(group.add(u, group.multiply(h1,c1))) &&
                 group.multiply(g2, z).equals(group.add(v, group.multiply(h2,c1))) &&
                 group.multiply(g1Tag, zTag).equals(group.add(uTag, group.multiply(h1Tag,c2))) &&
                 group.multiply(g2Tag, zTag).equals(group.add(vTag, group.multiply(h2Tag,c2)));

    }
}
